/*
	Usage of the functions in this module

	mem_init( num_base )
	...
	for each SZTN {
		mem_flush()
		for each path {
			mem_newpath()   // "N"
			for each node {
				mem_addnode( p_node )   // "%4d %4d (M|L)"
			}
			mem_setpath( closed )   // "(C|S)"
		}
		mem_alias()
		mem_rsort()
		mem_tonemerge()

	}
	...
	mem_fin()
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include "dptm2.h"

intCount ary_nbase;    // ary2[0] length
intCount ary_ndata;    // data length (# of nodes)
intCount ary_maxlen;   // ary2 allocated length
intCount ary_nlevel;   // ary2 allocated level
intCount ary_maxlev;   // ary2 maximum level
intCount seek1;   //
intCount seek2;   // these refer new (empty) ary2[][]
s_Node* p_startnode;
s_Node* p_prevnode;
s_Node** ary2;   // length = { nbase, nbase*1, nbase*2, nbase*4, ... }
s_Node** ary;    // access to ary2 as continuous array

extern FILE* fplog;

intCount seek2ndata( intCount seek1, intCount seek2 )
{
	/* ((1<<seek1)/2) = { 0, 1, 2, 4, 8, ... } */
	return ary_nbase * ( ( 1 << seek1 ) / 2 ) + seek2;
}

BOOL samepoint( s_Node* n1, s_Node* n2 )
{
	return ( n1->x == n2->x && n1->y == n2->y );
}

void delete_node( s_Node* p_node )
{
	/* 
	It is allowed to access the members except prev
	after delete the node.
	*/
	p_node->prev = p_node;   // disconnect
	return;
}

BOOL node_deleted( s_Node* p_node )
{
	assert( p_node != NULL );
	return ( p_node->prev == p_node );
}

void fputs_node( FILE* fp, s_Node* p_node, char op[4] )
{
	fprintf( fp, "%4d %4d %s\n", p_node->x, p_node->y, op );
	delete_node( p_node );
	return;
}

s_Node* alloc_pow2( intCount nlev )
{
	/* ary2[nlev] = alloc_pow2( max( nlev, 1 ) ) */
	size_t size = sizeof(s_Node);
	s_Node* ary_tmp = NULL;

	assert( ary_maxlev > 0 );
	assert( nlev > 0 );

	/* Array length must not be greater than MEMLENMAX. */
	if( nlev >= ary_maxlev )
	{
		fprintf( stderr, "Error: Reached the memory limit.\n" );
		exit( EXIT_FAILURE );
	}
	ary_tmp = (s_Node*)malloc( size * ary_nbase * ( 1 << (nlev-1) ) );   // not initialized
	//ary_tmp = (s_Node*)calloc( ary_nbase * ( 1 << (nlev-1) ), size );   // zero-cleared

	/* Check whether the memory was allocated successfully. */
	if( ary_tmp == NULL )
	{
		fprintf( stderr, "Error: Failed to allocate new memory in alloc_pow2.\n" );
		exit( EXIT_FAILURE );
	}

	return ary_tmp;
}

/* expand ary2 */
int expand_ary2( void )
{
	assert( 1 <= ary_nlevel && ary_nlevel < ary_maxlev );
	assert( ary == NULL );

	ary2[ary_nlevel] = alloc_pow2( ary_nlevel );
	ary_maxlen += ary_nbase * ( 1 << (ary_nlevel-1) );
	ary_nlevel++;

	chprintf( 'M', fplog, "Reallocate memory.  SUM = %d Bytes (%d nodes)\n", 
	         ary_maxlen * sizeof(s_Node), ary_maxlen );

	return 0;
}

int mem_init( intCount num_base )
{
	intCount nblock;

	assert( 0 < num_base && num_base <= MEMLENMAX );

	ary_nbase = num_base;
	for( ary_maxlev = 0, nblock = MEMLENMAX / ary_nbase; nblock > 0 ; ary_maxlev++, nblock /= 2 );
	assert( ary_maxlev > 0 );   // same to the first assertion
	ary2 = (s_Node**)calloc( ary_maxlev, sizeof(s_Node**) );
	if( ary2 == NULL )
	{
		fprintf( stderr, "Error: Failed to allocate ary2.\n" );
		return 1;
	}
	ary2[0] = alloc_pow2( 1 );
	ary_maxlen = ary_nbase;
	ary_nlevel = 1;

	chprintf( 'M', fplog, "Allocate memory.  SUM = %d Bytes (%d nodes)\n", 
	         ary_maxlen * sizeof(s_Node), ary_maxlen );

	return 0;
}

/* erase tone data, but don't deallocate ary2 */
int mem_flush( void )
{
	/* delete alias array */
	free2( ary );

	/* set zero */
	ary_ndata = 0;
	seek1 = 0;
	seek2 = 0;

	return 0;
}

/* call if buf is "N" */
int mem_newpath( void )
{
	p_startnode = NULL;
	p_prevnode  = NULL;

	return 0;
}

/* add node data */
int mem_addnode( s_Node* p_node_in )
{
	s_Node* p_node;   // alias to ary2[][]

	assert( 0 <= seek1 && seek1 < ary_nlevel );
	assert( 0 <= seek2 && seek2 <= ary_nbase * ( ( ( 1 << seek1 ) + 1 ) / 2 ) );
	assert( ary_ndata == seek2ndata( seek1, seek2 ) );
	assert( ary == NULL );

	/* adjust seek1&2 */
	if( seek2 == ary_nbase * ( ( ( 1 << seek1 ) + 1 ) / 2 ) )
	{
		seek1++;
		seek2 = 0;
		if( seek1 == ary_nlevel )
		{
			/* allocate new array */
			expand_ary2();
		}
	}
	assert( ary_ndata == seek2ndata( seek1, seek2 ) );   // check consistency

	/* copy point data (prev & next are undef) */
	p_node  = &ary2[seek1][seek2];
	*p_node = *p_node_in;

	/* connect to previous node */
	if( p_prevnode != NULL )
	{
		p_node->prev = p_prevnode;
		p_node->prev->next = p_node;
	}
	else
	{
		p_node->prev = NULL;
		p_startnode = p_node;   // registrate
	}
	p_prevnode = p_node;   // update

	/* increment */
	seek2++;
	ary_ndata++;

	return 0;
}

/* call if buf is "C"(->closed=true) or "S"(->closed=false) */
int mem_setpath( BOOL closed )
{
	s_Node* p_node;

	if( p_startnode == NULL ) return 0;   // no nodes

	p_node = p_prevnode;

	if( closed )
	{
		/* connect to startnode */
		p_node->next = p_startnode;
		p_startnode->prev = p_node;
	}
	else
	{
		/* nextnode is NULL */
		p_node->next = NULL;
	}

	/* reset */
	mem_newpath();

	return 0;
}

/* make continuous array */
intCount mem_alias( void )
{
	intCount i, j, k;

	assert( 0 <= seek1 && seek1 < ary_nlevel );
	assert( 0 <= seek2 && seek2 <= ary_nbase * ( ( ( 1 << seek1 ) + 1 ) / 2 ) );
	assert( ary_ndata == seek2ndata( seek1, seek2 ) );
	assert( ary == NULL );

	/* allocate */
	ary = (s_Node**)malloc( sizeof(s_Node*) * ary_ndata );
	if( ary == NULL )
	{
		fprintf( stderr, "Error: Failed to allocate the alias array.\n" );
		exit( EXIT_FAILURE );
	}

	/* refer ary2[][] */
	k = 0;
	for( j = 0; j < seek1; j++ )
	{
		for( i = 0; i < ary_nbase*(((1<<j)+1)/2); i++ )
		{
			ary[k++] = &ary2[j][i];
		}
	}
	j = seek1;
	{
		for( i = 0; i < seek2; i++ )
		{
			ary[k++] = &ary2[j][i];
		}
	}

	assert( k == ary_ndata );   // check consistency

	return ary_ndata;
}

/* radix sort (bucket sort for x & y) */
int mem_rsort( void )
{
	intCount i;
	intCount xoff[NUM_BUCKET] = { 0 };   // zero-cleared
	intCount yoff[NUM_BUCKET] = { 0 };   // zero-cleared
	intCount xsum, ysum;
	s_Node* p_node;
	s_Node** ary_temp;

	/* make offset tables */
	for( i = 0; i < ary_ndata; i++ )
	{
		assert( 0 <= ary[i]->x && ary[i]->x < NUM_BUCKET );
		assert( 0 <= ary[i]->y && ary[i]->y < NUM_BUCKET );
		xoff[ary[i]->x]++;
		yoff[ary[i]->y]++;
	}
	xsum = 0;
	ysum = 0;
	for( i = 0; i < NUM_BUCKET; i++ )
	{
		xsum += xoff[i];   xoff[i] = xsum - xoff[i];
		ysum += yoff[i];   yoff[i] = ysum - yoff[i];
	}
	assert( xsum == ary_ndata );
	assert( xsum == ary_ndata );

	/* sort : ary -(y)-> ary_temp -(x)-> ary */
	ary_temp = (s_Node**)malloc( sizeof(s_Node*) * ary_ndata );
	if( ary_temp == NULL )
	{
		fprintf( stderr, "Error: Failed to allocate ary_temp in mem_rsort.\n" );
		exit( EXIT_FAILURE );
	}
	for( i = 0; i < ary_ndata; i++ )
	{
		p_node = ary[i];
		ary_temp[yoff[p_node->y]++] = p_node;
		assert( yoff[p_node->y] <= yoff[p_node->y+1] );
	}
	for( i = 0; i < ary_ndata; i++ )
	{
		p_node = ary_temp[i];
		ary[xoff[p_node->x]++] = p_node;
		assert( xoff[p_node->x] <= xoff[p_node->x+1] );
	}
	free( ary_temp );

	return 0;
}

intCount mem_tonemerge( void )
{
	intCount i, j;
	intCount k1, k2;
	intCount num_nodes;

	assert( ary != NULL );

	num_nodes = ary_ndata;
	for( i = 0; i < ary_ndata; i++ )
	{
		/* search same point nodes */
		for( j = i+1; j < ary_ndata; j++ )
		{
			if( !samepoint( ary[i], ary[j] ) ) break;
		}

		/* merge ary[k1] & ary[k2] (i<=k1<k2<j) */
		for( k1 = i; k1 < j-1; k1++ )
		{
			for( k2 = k1+1; k2 < j; k2++ )
			{
				if( samepoint( ary[k1]->next, ary[k2]->prev ) )
				{
					/* delete ary[k1] */
					/*
					ary[k1]->prev->next = ary[k2];
					ary[k2]->prev = ary[k1]->prev;
					delete_node( ary[k1] );
					num_nodes--;   // count
					break;
					*/

					/* connect (not delete) */
					ary[k1]->next->prev = ary[k2]->prev;
					ary[k2]->prev->next = ary[k1]->next;
					ary[k1]->next = ary[k2];
					ary[k2]->prev = ary[k1];
					num_nodes--;   // count
					break;
				}
			}
		}

		/* update i */
		i = j-1;
	}

	return num_nodes;
}

intCount mem_pathopt( void )
{
	intCount i;
	intCount count;

	assert( ary != NULL );

	count = 0;
	for( i = 0; i < ary_ndata; i++ )
	{
		intPoint d1x, d1y, d2x, d2y;   /* check not to overflow */

		if( node_deleted( ary[i] ) ) continue;

		assert( ary[i]->prev != NULL && ary[i]->next != NULL );
		d1x = ary[i]->prev->x - ary[i]->x;
		d1y = ary[i]->prev->y - ary[i]->y;
		d2x = ary[i]->next->x - ary[i]->x;
		d2y = ary[i]->next->y - ary[i]->y;
		if( d1x * d2y == d1y * d2x )
		{
			/* delete ary[i] */
			ary[i]->prev->next = ary[i]->next;
			ary[i]->next->prev = ary[i]->prev;
			delete_node( ary[i] );
		}
		else
		{
			count++;
		}
	}

	return count;
}

intCount mem_output_tone( FILE* fp )
{
	intCount i;
	intCount count;

	assert( ary != NULL );

	count = 0;
	fprintf( fp, "N\n" );   // newpath
	for( i = 0; i < ary_ndata; i++ )
	{
		s_Node* p_node;

		/* skip */
		if( node_deleted( ary[i] ) ) continue;

		fputs_node( fp, ary[i], "M" );   // start a subpath
		count++;
		for( p_node = ary[i]->next; p_node != ary[i]; p_node = p_node->next )
		{
			fputs_node( fp, p_node, "L" );
			count++;
		}
		fprintf( fp, "C\n" );   // closepath
	}


	return count;
}

intCount mem_output_line( FILE* fp ) /* v1.0: don't use rsort! */
{
	intCount i;
	intCount count;

	assert( ary != NULL );

	count = 0;
	for( i = 0; i < ary_ndata; i++ )
	{
		if( ary[i]->prev == NULL )
		{
			fputs( "N\n", fp );   /* newpath */
		}

		fputs_node( fp, ary[i], ary[i]->op );
		count++;

		if( ary[i]->next == NULL )
		{
			fputs( "S\n", fp );   /* stroke */
		}
	}

	return count;
}

int mem_fin( void )
{
	intCount i;

	free2( ary );

	for( i = 0; i < ary_maxlev; i++ )
	{
		assert( i < ary_nlevel || ary2[i] == NULL );
		free2( ary2[i] );
	}
	ary_maxlen = 0;

	chprintf( 'M', fplog, "Deallocate memory.\n" );

	return 0;
}
